@@ -11,7 +11,7 @@
   (*  Contact: dev@gim.name                                                 *)
   (**************************************************************************)
 
-(* $Id: dtools.ml 4682 2007-10-17 19:24:03Z metamorph68 $ *)
+(* $Id: dtools.ml 4940 2007-12-12 12:44:59Z metamorph68 $ *)
 
 (**
   ocaml-dtools
@@ -535,23 +535,36 @@ struct
     end
 
   let daemonize fn =
-    close_in stdin;
-    flush_all ();
     begin match Unix.fork () with
     | 0 ->
-	let _ = Unix.setsid () in
-	begin match conf_daemon_pidfile#get with
-	| false ->
-	    catch fn (fun () -> ())
-	| true ->
-	    let filename = conf_daemon_pidfile_path#get in
-	    let f = open_out filename in
-	    let pid = Unix.getpid () in
-	    output_string f (string_of_int pid);
-	    output_char f '\n';
-	    close_out f;
-	    catch fn (fun () -> Unix.unlink filename)
-	end;
+        (* Change umask to 0 *)
+        ignore(Unix.umask 0) ;
+	let f =  
+          match conf_daemon_pidfile#get with
+	    | false ->
+	       fun () -> catch fn (fun () -> ())
+	    | true ->
+	       (* Write PID to file *)
+	       let filename = conf_daemon_pidfile_path#get in
+	       let f = open_out filename in
+	       let pid = Unix.getpid () in
+	       output_string f (string_of_int pid);
+	       output_char f '\n';
+	       close_out f ;
+               fun () -> catch fn (fun () -> Unix.unlink filename)
+	in
+        (* Dettach from the console *)
+        if (Unix.setsid () < 0) then
+               exit 1;
+        (* chdir to / *)
+        Unix.chdir "/" ;
+        (* Close usual file descriptor *)
+        flush_all ();
+        close_in stdin;
+        close_out stdout;
+        close_out stderr;
+        (* Main loop *)
+        f () ;
         exit 0
     | _ -> exit 0
     end
@@ -559,9 +572,9 @@ struct
   let exit_when_root () =
     let security s = Printf.eprintf "init: security exit, %s\n%!" s in
     if Unix.geteuid () = 0 then
-      begin security "root euid."; exit (-1) end;
+      begin security "root euid (user)."; exit (-1) end;
     if Unix.getegid () = 0 then
-      begin security "root egid."; exit (-1) end
+      begin security "root egid (group)."; exit (-1) end
 
   let init ?(prohibit_root=false) f =
     if prohibit_root then exit_when_root ();
@@ -698,14 +711,22 @@ struct
       method active lvl =
 	let rec aux l =
 	  begin match l with
-	  | [] -> false
+	  | [] -> None
 	  | t :: q ->
-	      begin try lvl <= (Conf.as_int t)#get with
-	      | Conf.Undefined _ -> aux q
+	      begin match aux q with
+	      | Some i -> Some i
+	      | None ->
+		  begin try Some ((Conf.as_int t)#get)
+		  with
+		  | Conf.Undefined _ -> None
+		  end
 	      end
 	  end
 	in
-	aux confs
+	begin match aux confs with
+	| None -> false
+	| Some i -> i >= lvl
+	end
       method f lvl =
 	begin match self#active lvl with
 	| true ->
